@@ -29,8 +29,8 @@
 /* ----------------------------------
  * version info
  * ---------------------------------- */
-static const char man_version[] = "11.2";
-static const char man_date[] = "May 2014";
+static const char man_version[] = "11.5";
+static const char man_date[] = "August 2014";
 
 /* ----------------------------------
  * definitions
@@ -40,17 +40,6 @@ static const char man_date[] = "May 2014";
 #define CR 13
 #define LF 10
 
-/* ----------------------------------
- * structures
- * ---------------------------------- */
-typedef struct qs_event_st {
-  char       *id;    /**< id, e.g. ip address or client correlator string */
-  time_t     time;   /**< last update, used for expiration */
-  int        count;  /**< event count/updates */
-  struct qs_event_st *next;
-} qs_event_t;
-
-
 /* ----------------------------------
  * functions
  * ---------------------------------- */
@@ -72,13 +61,6 @@ void qs_csInitLock();
 void qs_csLock();
 void qs_csUnLock();
 
-/* events */
-void qs_setExpiration(time_t sec);
-int  qs_insertEvent(qs_event_t **l_qs_event, char *id);
-long qs_countEvent(qs_event_t **l_qs_event);
-void qs_deleteEvent(qs_event_t **l_qs_event, char *id);
-void qs_GCEvent(qs_event_t **l_qs_event);
-
 /* log */
 void qs_deleteOldFiles(const char *file_name, int generations);
 
