@@ -28,7 +28,7 @@
  *
  */
 
-static const char revision[] = "$Id: qslog.c,v 1.84 2014/04/30 13:19:41 pbuchbinder Exp $";
+static const char revision[] = "$Id: qslog.c,v 1.95 2014/07/09 20:02:05 pbuchbinder Exp $";
 
 #include <stdio.h>
 #include <string.h>
@@ -58,9 +58,12 @@ static const char revision[] = "$Id: qslog.c,v 1.84 2014/04/30 13:19:41 pbuchbin
  * ---------------------------------- */
 #define ACTIVE_TIME 600 /* how long is a client "active" (ip addresses seen in the log) */
 #define LOG_INTERVAL 60 /* log interval ist 60 sec, don't change this value */
+#define QS_GC_INTERVAL 10
 #define LOG_DET ".detailed"
 #define RULE_DELIM ':'
 #define MAX_CLIENT_ENTRIES 25000
+#define MAX_EVENT_ENTRIES  50000
+#define NUM_EVENT_TABLES   8
 #define QS_GENERATIONS 14
 #define EVENT_DELIM ','
 #define QSEVENTPATH "QSEVENTPATH" /* varibale name to find event definitions */
@@ -158,14 +161,25 @@ typedef struct stat_rec_st {
   apr_pool_t *pool;
 } stat_rec_t;
 
+typedef struct qs_event_st {
+  char       *id;    /**< id, e.g. ip address or client correlator string */
+  time_t     time;   /**< last update, used for expiration */
+  long       count;  /**< event count/updates */
+} qs_event_t;
+
 /* ----------------------------------
  * global stat counter
  * ---------------------------------- */
 static stat_rec_t* m_stat_rec;
 static stat_rec_t* m_stat_sub = NULL;
 
-static qs_event_t *m_ip_list = NULL;
-static qs_event_t *m_user_list = NULL;
+static time_t m_qs_expiration = 60 * 10;
+
+static apr_table_t *m_ip_list[NUM_EVENT_TABLES];   /* IP session store */
+static apr_table_t *m_user_list[NUM_EVENT_TABLES]; /* user session store */
+static int m_log_max = 0;                          /* already reached store limit */
+static int m_hasGC = 0;                            /* sep. gc thread or not */
+static qs_event_t **m_gc_event_list = NULL;        /* list of entries to delete */
 
 /* output file */
 static FILE *m_f = NULL;
@@ -197,6 +211,187 @@ static int   m_methods = 0;
 static long  m_lines = 0;
 static int   m_verbose = 0;
 
+/* events ----------------------------------------------------- */
+/*
+ * sets the expiration for events
+ */
+void qs_setExpiration(time_t sec) {
+  m_qs_expiration = sec;
+}
+
+/*
+ * creates a new event entry
+ */
+static qs_event_t *qs_newEvent(const char *id) {
+  qs_event_t *ev = calloc(sizeof(qs_event_t), 1);
+  ev->id = calloc(strlen(id) + 1, 1);
+  strcpy(ev->id, id);
+  qs_time(&ev->time);
+  ev->count = 1;
+  return ev;
+}
+
+/*
+ * deletes an event
+ */
+void qs_freeEvent(qs_event_t *ev) {
+  free(ev->id);
+  free(ev);
+}
+
+/**
+ * Defines in which of the NUM_EVENT_TABLES session stores
+ * the id shall be added.
+ *
+ * @param str Session ID to store
+ * @return The id of the storage table (0 <= n < NUM_EVENT_TABLES)
+ */
+static int qs_tableSelector(const char *str) {
+  int num = 0;
+  int len = strlen(str);
+  if(len > 3) {
+    if(str[len-1] == '=' ||
+       str[len-1] == '\'' ||
+       str[len-1] == '"') {
+      len--;
+    }
+  }
+  if(str[0] % 2 == 1) {
+    num += 1;
+  }
+  if(len > 1) {
+    if(str[len-1] % 2 == 1) {
+      num += 2;
+    }
+    if(len > 2) {
+      if(str[len-2] % 2 == 1) {
+        num += 4;
+      }
+    }
+  }
+  return num;
+}
+
+/**
+ * Inserts an event entry and delets expired.
+ *
+ * @param l_qs_event Pointer to the event list.
+ * @param id Identifer, e.g. IP address or user tracking cookie
+ *
+ * @return event counter (number of updates) for the provided id
+ */
+static long qs_insertEventT(apr_table_t **list0, const char *id) {
+  qs_event_t *lp;
+  int select = qs_tableSelector(id);
+  apr_table_t *list = list0[select];
+  lp = (qs_event_t *)apr_table_get(list, id);
+  if(lp) {
+    // exists
+    qs_time(&lp->time);
+    lp->count++;
+    return lp->count;
+  }
+  if(apr_table_elts(list)->nelts >= MAX_EVENT_ENTRIES) {
+    if(m_log_max == 0) {
+      char time_string[1024];
+      time_t tm = time(NULL);
+      struct tm *ptr = localtime(&tm);
+      strftime(time_string, sizeof(time_string), "%a %b %d %H:%M:%S %Y", ptr);
+      fprintf(stderr, "[%s] [notice] qslog: reached event count limit\n",
+              time_string);
+      m_log_max = 1;
+    }
+    return 0;
+  }
+  lp = qs_newEvent(id);
+  apr_table_setn(list, lp->id, (char *)lp);
+  return lp->count;
+}
+
+/**
+ * deletes expired events
+ */
+static void gcTable(apr_table_t *list) {
+  int max = 0;
+  int i;
+  apr_table_entry_t *entry;
+  time_t gmt_time;
+  qs_time(&gmt_time);
+
+  if(m_hasGC) {
+    qs_csLock();
+  }
+  // collect expired events...
+  entry = (apr_table_entry_t *) apr_table_elts(list)->elts;
+  for(i = 0; i < apr_table_elts(list)->nelts; i++) {
+    qs_event_t *lp = (qs_event_t *)entry[i].val;    
+    if(lp->time < (gmt_time - m_qs_expiration)) {
+      m_gc_event_list[max] = lp;
+      max++;
+    }
+  }
+  // ...remove...
+  for(i = 0; i < max; i++) {
+    if(m_hasGC) {
+      /* we don't want to hold a lock for a long time
+         => temp release the lock letting the pipe-buffer recover */
+      if(i % 10 == 9) {
+        qs_csUnLock();
+        // wait 1ms
+        apr_sleep(1000);
+        qs_csLock();
+      }
+    }
+    apr_table_unset(list, m_gc_event_list[i]->id);
+  }
+  if(m_hasGC) {
+    qs_csUnLock();
+  }
+
+  // ...and delete them
+  for(i = 0; i < max; i++) {
+    qs_freeEvent(m_gc_event_list[i]);
+  }
+}
+
+/**
+ * Returns the number of events
+ *
+ * @param event table
+ * @return Number of entries
+ */
+static long qs_countEventT(apr_table_t **list) {
+  int count = 0;
+  int t;
+  if(!m_hasGC) {
+    for(t = 0; t < NUM_EVENT_TABLES; t++) {
+      gcTable(list[t]);
+    }
+  }
+  for(t = 0; t < NUM_EVENT_TABLES; t++) {
+    count += apr_table_elts(list[t])->nelts;
+  }
+  return count;
+}
+
+/**
+ * Calls the event table GC
+ */
+static void *gcThread(void *argv) {
+  int t;
+  m_hasGC = 1;
+  while(1) {
+    sleep(QS_GC_INTERVAL);
+    for(t = 0; t < NUM_EVENT_TABLES; t++) {
+      gcTable(m_ip_list[t]);
+      gcTable(m_user_list[t]);
+    }
+  }
+  return NULL;
+}
+
+/* ------------------------------------------------------------ */
+
 /**
  * Helper to print an error message when terminating
  * the programm due to an unexpected error.
@@ -491,8 +686,8 @@ static void printStat2File(FILE *f, char *timeStr, stat_rec_t *stat_rec,
             stat_rec->averAge / (stat_rec->averAge_count == 0 ? 1 : stat_rec->averAge_count));
   }
   if(main) {
-    sprintf(ip, "ip;%ld;", qs_countEvent(&m_ip_list));
-    sprintf(usr, "usr;%ld;", qs_countEvent(&m_user_list));
+    sprintf(ip, "ip;%ld;", qs_countEventT(m_ip_list));
+    sprintf(usr, "usr;%ld;", qs_countEventT(m_user_list));
   } else {
     ip[0] = '\0';
     usr[0] = '\0';
@@ -1053,11 +1248,11 @@ static void updateRec(stat_rec_t *rec, char *T, char *t, char *D, char *S,
   }
   if(I != NULL) {
     /* update/store client IP */
-    qs_insertEvent(&m_ip_list, I);
+    qs_insertEventT(m_ip_list, I);
   }
   if(U != NULL) {
     /* update/store user */
-    qs_insertEvent(&m_user_list, U);
+    qs_insertEventT(m_user_list, U);
   }
   if(B != NULL) {
     /* transferred bytes */
@@ -1534,7 +1729,7 @@ static void *loggerThread(void *argv) {
     strftime(buf, sizeof(buf), "%d.%m.%Y %H:%M:%S", ptr);
 
     printAndResetStat(buf);
-    if(m_rotate) {
+    if(m_rotate && m_file_name[0]) {
       strftime(buf, sizeof(buf), "%H:%M", ptr);
       if(strcmp(buf, "23:59") == 0) {
         char arch[MAX_LINE];
@@ -1646,7 +1841,8 @@ static void usage(const char *cmd, int man) {
   if(man) printf("\n.TP\n");
   qs_man_print(man, "  -o <out_file>\n");
   if(man) printf("\n");
-  qs_man_print(man, "     Specifies the file to store the output to.\n");
+  qs_man_print(man, "     Specifies the file to store the output to. stdout is used if this option\n");
+  qs_man_print(man, "     is not defined.\n");
   if(man) printf("\n.TP\n");
   qs_man_print(man, "  -p\n");
   if(man) printf("\n");
@@ -1693,16 +1889,7 @@ static void usage(const char *cmd, int man) {
   } else {
     printf("Example configuration using pipped logging:\n");
   }
-  qs_man_println(man, "  LogFormat \"%%t %%h \\\"%%r\\\" %%>s %%b \\\"%%{User-Agent}i\\\" %%T\"\n");
-  qs_man_println(man, "  TransferLog \"|/bin/%s -f ..IRSB.T -x -o /var/logs/stat_log\"\n", cmd);
-  printf("\n");
-  if(man) {
-    printf("Configuration using the CustomLog directive:\n");
-    printf("\n");
-  } else {
-   printf("Example configuration using the CustomLog directive:\n");
-  }
-  qs_man_println(man, "  CustomLog \"|/bin/%s -f ISBTQ -x -o /var/logs/stat_log\" \"%%h %%>s %%b %%T %%{mod_qos_ev}e\"\n", cmd);
+  qs_man_println(man, "  CustomLog \"|/bin/%s -f ISBTQ -x -o /var/logs/stat.csv\" \"%%h %%>s %%b %%T %%{mod_qos_ev}e\"\n", cmd);
   printf("\n");
   if(man) {
     printf("Post processing:\n");
@@ -1710,7 +1897,8 @@ static void usage(const char *cmd, int man) {
   } else {
     printf("Example for post processing:\n");
   }
-  qs_man_println(man, "  cat access_log | /bin/%s -f ..IRSB.T -o /var/logs/stat_log -p\n", cmd);
+  qs_man_println(man, "  LogFormat \"%%t %%h \\\"%%r\\\" %%>s %%b \\\"%%{User-Agent}i\\\" %%T\"\n");
+  qs_man_println(man, "  cat access_log | /bin/%s -f ..IRSB.T -o /var/logs/stat.csv -p\n", cmd);
   printf("\n");
   if(man) {
     printf(".SH SEE ALSO\n");
@@ -1780,11 +1968,20 @@ int main(int argc, const char *const argv[]) {
   const char *username = NULL;
   pthread_attr_t *tha = NULL;
   pthread_t tid;
+  pthread_attr_t *thagc = NULL;
+  pthread_t tidgc;
   apr_pool_t *pool;
+  int t;
   apr_app_initialize(&argc, &argv, NULL);
   apr_pool_create(&pool, NULL);
   m_stat_rec = createRec(pool, "", "");
 
+  for(t = 0; t < NUM_EVENT_TABLES; t++) {
+    m_ip_list[t] = apr_table_make(pool, 15000);
+    m_user_list[t] = apr_table_make(pool, 15000);
+  }
+  m_gc_event_list = calloc(MAX_EVENT_ENTRIES, sizeof(qs_event_t *));
+
   qs_csInitLock();
   qs_setExpiration(ACTIVE_TIME);
   if(cmd == NULL) {
@@ -1903,6 +2100,7 @@ int main(int argc, const char *const argv[]) {
     long long duration_count_ms_min = -1;
     long long duration_count_ms_max = 0;
     m_url_entries = apr_table_make(pool, MAX_CLIENT_ENTRIES + 1);
+    if(config == NULL) usage(cmd, 0);
     readStdinOffline(pool, config);
     fprintf(stderr, ".\n");
 
@@ -1945,7 +2143,7 @@ int main(int argc, const char *const argv[]) {
               url_rec->status_3,
               url_rec->status_4,
               url_rec->status_5,
-              url_rec->duration_count_ms / url_rec->request_count,
+              url_rec->request_count ? (url_rec->duration_count_ms / url_rec->request_count) : 0,
               entry[i].key);
       
     }
@@ -1962,8 +2160,8 @@ int main(int argc, const char *const argv[]) {
     fprintf(m_f, "req;%ld;"
             ";;;;;;;;;;"
             NAVMS";%lld;average;\n",
-            request_count / apr_table_elts(m_url_entries)->nelts,
-            duration_count_ms / request_count);
+            apr_table_elts(m_url_entries)->nelts ? (request_count / apr_table_elts(m_url_entries)->nelts) : 0,
+            request_count ? (duration_count_ms / request_count) : 0);
     if(file && m_f != stdout) {
       fclose(m_f);
     }
@@ -1998,8 +2196,8 @@ int main(int argc, const char *const argv[]) {
       /* ci (coverage index): low value indicates that we have seen the client
                               at the end or beginning of the file (maybe not all
                               requests due to log rotation) */
-      long coverage = (client_rec->firstLine * 100 / m_lines);
-      long coverageend = 100 - ((client_rec->lastLine * 100) / m_lines);
+      long coverage = m_lines ? (client_rec->firstLine * 100 / m_lines) : 0;
+      long coverageend = m_lines ? (100 - ((client_rec->lastLine * 100) / m_lines)) : 0;
       if(coverageend < coverage) {
         coverage = coverageend;
       }
@@ -2037,8 +2235,8 @@ int main(int argc, const char *const argv[]) {
               client_rec->status_4,
               client_rec->status_5,
               client_rec->status_304,
-              client_rec->duration / client_rec->request_count,
-              client_rec->duration_count_ms / client_rec->request_count,
+              client_rec->request_count ? (client_rec->duration / client_rec->request_count) : 0,
+              client_rec->request_count ? (client_rec->duration_count_ms / client_rec->request_count) : 0,
               client_rec->duration_0,
               client_rec->duration_1,
               client_rec->duration_2,
@@ -2068,14 +2266,14 @@ int main(int argc, const char *const argv[]) {
       }
       fprintf(m_f, "\n");
     }
-    if(file && m_f != stdout) {
+    if(file && (m_f != stdout)) {
       fclose(m_f);
     }
     return 0;
   }
 
-  /* requires at least an output file and a format string */
-  if(file == NULL || config == NULL) usage(cmd, 0);
+  /* requires at least a format string */
+  if(config == NULL) usage(cmd, 0);
 
   if(username && getuid() == 0) {
     struct passwd *pwd = getpwnam(username);
@@ -2098,18 +2296,27 @@ int main(int argc, const char *const argv[]) {
     }
   }
 
-  m_f = fopen(file, "a+"); 
-  if(m_f == NULL) {
-    qerror("could not open file for writing '%s': %s", file, strerror(errno));
-    exit(1);
+  if(file) {
+    m_f = fopen(file, "a+"); 
+    if(m_f == NULL) {
+      qerror("could not open file for writing '%s': %s", file, strerror(errno));
+      exit(1);
+    }
+    if(strlen(file) > (sizeof(m_file_name) - strlen(".yyyymmddHHMMSS  ") - strlen(LOG_DET))) {
+      qerror("file name too long '%s'", file);
+      exit(1);
+    }
+    strcpy(m_file_name, file);
+  } else {
+    m_file_name[0] = '\0';
+    m_f = stdout;
   }
-  if(strlen(file) > (sizeof(m_file_name) - strlen(".yyyymmddHHMMSS  ") - strlen(LOG_DET))) {
-    qerror("file name too long '%s'", file);
-    exit(1);
-  }
-  strcpy(m_file_name, file);
 
   if(confFile) {
+    if(file == NULL) {
+      qerror("option '-c' can only be used in conjunction with option '-o'");
+      exit(1);
+    }
     snprintf(m_file_name2, sizeof(m_file_name2), "%s"LOG_DET, m_file_name);
     if(strchr(config, 'C') == NULL) {
       qerror("you need to add 'C' to the format string when enabling the pattern list (-c)");
@@ -2124,12 +2331,12 @@ int main(int argc, const char *const argv[]) {
   }
 
   /*
-   * Offline mode reads an existing file
+   * Offline mode reads an existing log file
    * adjusting a virtual clock based on
    * the date string match of the log
    * enties. */
   if(m_offline) {
-    fprintf(stderr, "[%s]: offline mode (writes to %s)\n", cmd, file);
+    fprintf(stderr, "[%s]: offline mode\n", cmd);
     m_date_str[0] = '\0';
     readStdinOffline(pool, config);
     if(!m_verbose) {
@@ -2141,8 +2348,11 @@ int main(int argc, const char *const argv[]) {
      * to write the data every minute. 
      */
     pthread_create(&tid, tha, loggerThread, NULL);
+    pthread_create(&tidgc, thagc, gcThread, NULL);
     readStdin(pool, config);
   }
-  fclose(m_f);
+  if(file && (m_f != stdout)) {
+    fclose(m_f);
+  }
   return 0;
 }
